/*
    Str912 book EEPROM emulation library
    Copyright (C) 2007  Lucjan Bryndza <lucjan.bryndza@ep.com.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <91x_fmi.h>
#include <stdbool.h>
#include <stdint.h>
#include "eeprom.h"

/*---------------------------------------------------------------------*/
//Sector FMI_B1S0 Bank FMI_BANK1

static void fmiWriteProtectionCmd(vu32 sector, bool enable)
{
  if (enable == true)
  {
    *(vu16*)sector = 0x60;
    *(vu16*)sector = 0x01;
    *(vu16*)sector = 0xFF;
  }
  else
  {
    *(vu16*)sector = 0x60;
    *(vu16*)sector = 0xD0;
    *(vu16*)sector = 0xFF;
  }
}


/*---------------------------------------------------------------------*/
static void fmiEraseSector(vu32 sector)
{
  //Komenda ustawienia kasowania
  *(vu16 *)sector = 0x20;

  // Komenda potwierdzenie kasowania sektora
  *(vu16 *)sector = 0xD0;
}

/*---------------------------------------------------------------------*/
static int fmiWaitForLastOperation(vu32 bank)
{
  u32 tOut = 0;

  // Zapisz rejestr statusu
  *(vu16 *)(bank) = 0x70;

  //Czekaj do zakonczenia operacji
  while((!((*(vu16 *)bank) & 0x80))&&(tOut < 0xffff ))
  {
    tOut++;
  }

  /* Komenda read array */
  *(vu16 *)bank = 0xFF;
  if (tOut == 0xffff)
  {
    return -1;
  }
  else
  {
    return 0;
  }
}

/*---------------------------------------------------------------------*/

static void fmiWriteHalfWord(u32 addr, u16 data)
{
  /* Write a program command to the sector to be written */
  *(vu16 *)(addr & 0xFFFFFFFC) = 0x40;

  /* Write the halfword to the destination address */
  *(vu16 *)addr = data;
}

/*---------------------------------------------------------------------*/
#define fmiReadWord(addr) *(vu16*)(addr)

/*---------------------------------------------------------------------*/

#define BANK_ACTIVE 0x1979

/*---------------------------------------------------------------------*/
//Read eeprom data at specified address
int eepromflaRead(uint16_t addr)
{
    if(addr > EEPROM_MAX_VA) return EEPROMERR_ADDR;
    uint32_t bank = FMI_B1S0;
    if(fmiReadWord(bank)!=BANK_ACTIVE)
    {
	    bank = FMI_B1S1;
	    if(fmiReadWord(bank)!=BANK_ACTIVE) return EEPROMERR_INIT;
    }
    for(int i=EEPROM_MAX_PA;i>=4;i-=4)
    {
	    if( fmiReadWord(bank+i) == addr) return fmiReadWord(bank+i+2);
    }
    return EEPROMERR_ADDR;
}

/*---------------------------------------------------------------------*/
//Internal function to sort flash memory if memory ends.
static int sortFlash(void)
{
   uint32_t bActive = FMI_B1S0;
   uint32_t bNew = FMI_B1S1;

   if(fmiReadWord(bActive)!=BANK_ACTIVE)
   {
        bActive = FMI_B1S1;
        bNew = FMI_B1S0;
   }
   for(int va=0,paNew=4;va<EEPROM_MAX_VA;va++)
   {
       for(int pa=EEPROM_MAX_PA;pa>=4;pa-=4)
       {
            if(fmiReadWord(bActive+pa) == va)
            {
               if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
	           fmiWriteHalfWord((bNew+paNew),va);
	           if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
	           fmiWriteHalfWord((u32)(bNew+paNew+2),fmiReadWord(bActive+pa+2));
	           if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
            	   paNew += 4;
		   break;
            }
       }
   }
   //Kasuj aktywny sektor
   fmiEraseSector(bActive);
   if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
   //Zapisz status aktywnosci we wskazanym banku
   fmiWriteHalfWord(bNew,BANK_ACTIVE);
   if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
   return 0;
}

/*---------------------------------------------------------------------*/
//Write flash at specified virtual address
int eepromflaWrite(uint16_t addr,uint16_t data)
{
    if(addr > EEPROM_MAX_VA) return EEPROMERR_ADDR;
    uint32_t bank = FMI_B1S0;
    if(fmiReadWord(bank)!=BANK_ACTIVE)
    {
	    bank = FMI_B1S1;
	    if(fmiReadWord(bank)!=BANK_ACTIVE) return EEPROMERR_INIT;
    }
    for(int i=4;i<EEPROM_MAX_PA;i+=4)
    {
	    if( fmiReadWord(bank+i) == 0xffff )
	    {
	        if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
	        fmiWriteHalfWord((bank+i),addr);
	        if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
	        fmiWriteHalfWord((bank+i+2),data);
	        if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
            if(i==(EEPROM_MAX_PA-4)) return sortFlash();
            else return 0;
	    }
    }
    return EEPROMERR_NOSPACE;
}

/*---------------------------------------------------------------------*/
//Initialize eeprom procedure
int eepromflaInit()
{
    //Disable protection for sector 0
    fmiWriteProtectionCmd(FMI_B1S0, false);
    //Disable protection for sector 1
    fmiWriteProtectionCmd(FMI_B1S1, false);

    uint32_t bank = FMI_B1S0;
    if(fmiReadWord(bank)!=BANK_ACTIVE)
    {
	    bank = FMI_B1S1;
	    if(fmiReadWord(bank)!=BANK_ACTIVE)
        {
            fmiEraseSector(FMI_B1S0);
            if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
            fmiWriteHalfWord(FMI_B1S0,BANK_ACTIVE);
            if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
            fmiEraseSector(FMI_B1S1);
            if(fmiWaitForLastOperation(FMI_BANK_1)<0) return EEPROMERR_TIMEOUT;
        }
    }
    return 0;
}


