/*
    Str912 book example16 - ADC with interrupts in continous mode
    Copyright (C) 2007  Lucjan Bryndza <lucjan.bryndza@ep.com.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "91x_map.h"
#include "91x_scu.h"
#include "91x_fmi.h"
#include "91x_adc.h"
#include "armint.h"
#include "uart.h"
#include <stdio.h>
/*----------------------------------------------------------------------*/
//UART for output
#define USED_UART UART0

/*----------------------------------------------------------------------*/

// delay for n cycles
static void delay(int n)
{
    while(n--)
	asm volatile ("nop");
}

/*----------------------------------------------------------------------*/

/* Ustawienie parametrow zegarowych */
//Czestotliwosc 96MHz
#define PLL_N 192
#define PLL_M 25
#define PLL_P 2
#define SCU_PLLEN 0x80000

static void  setup_pll(void)
{
    //Jezeli petla jest zsynchronizowana tzn ze jest ok
    if(SCU->SYSSTATUS & SCU_FLAG_LOCK) return;
    //Konfiguracja uruchomienie i synchronizacja PLL
    SCU->PLLCONF = (PLL_P<<16) | (PLL_N<<8) | PLL_M;
    SCU->PLLCONF |= SCU_PLLEN;
    while(~SCU->SYSSTATUS & SCU_FLAG_LOCK);

    //Ustawiamy pamiec flash na dwa cykle oczekiwania (dla 96MHz)
    FMI->CR |= FMI_WRITE_WAIT_STATE_1;
    *(vu16*)FMI_BANK_1 = 0x60;
    *(vu16*)(FMI_BANK_1|FMI_READ_WAIT_STATE_2|FMI_LVD_ENABLE|FMI_PWD_ENABLE|FMI_FREQ_HIGH) = 0x03;

    //Ustawiamy F clk na koncu bedzie z PLL-a
    SCU->CLKCNTR = SCU_EMIBCLK_Div2 | SCU_USBCLK_MCLK2 | SCU_PCLK_Div2;
}


/*-------------------------------------------------------------*/
//Wartosc usredniona A/C

static uint16_t adc_val;

/*-------------------------------------------------------------*/
#define ADC_ECV (1<<15)
#define ADC_STR (1<<0)


//Przerwanie niemaskowalne fiq
static void adcIRQHandler(void) __attribute__ ((interrupt("IRQ")));


void adcIRQHandler(void)
{
    static uint32_t acc;
    static uint16_t pos;
    //Kasowanie flagi ECV
    ADC->CR |= ADC_ECV;
    acc+= ADC->DR0 & 0x3ff;
    if(++pos>=256)
    {
        adc_val = acc/256;
        acc = pos = 0;
    }
    //Informacja dla VIC (koniec przerwania)
    VIC0->VAR = 0;
}

/*---------------------------------------------------------------*/

/* Ustawianie przetwornika A/C w trybie ciaglej konwersji */

#define ADC_ENABLE (1<<1)
#define ADC_STANDBY (1<<3)
#define ADC_CHANNEL_0 (0<<6)
#define ADC_CC0_NOWDT (3<<0)
#define ADC_CONT 0x10
#define ADC_ECVI (1<<10)


#define VIC0_ADC (1<<15)
#define VIC0_ADC_BIT 15
#define VIC_SLOT_EN (1<<5)

static void setup_adc(void)
{
    //Wlacz sygnal zegarowy dla GPIO4 i ADC
    SCU->PCGR1 |= __GPIO4 | __ADC;
    //Odblokuj sygnal resetu dla GPIO04 i ADC
    SCU->PRR1 |= __GPIO4 | __ADC;
    //Dolacz wejscie P4.0 do ADC
    SCU->GPIOIN[4] |= (1<<0);
    //Wlacz tryb analogowy linii P4.0
    SCU->GPIOANA |= (1<<0);
    //Ustaw preskaler czestoliwosc musi byc >25MHz
    ADC->PRS = 20;
    //Enable ADC and select channel 0
    ADC->CR = ADC_ENABLE;
    ADC->CR |= ADC_CHANNEL_0;
    //Ustawienie bez watchdoga analogowego na P0.4
    ADC->CCR |= ADC_CC0_NOWDT;

    //Zalaczenie clocka dla VIC
    SCU->PCGR0 |= __VIC;
    //Zwolnienie ewentualnego resetu
    SCU->PRR0 |= __VIC;
    //Do wektora podstaw adres procedury
    VIC0->VAiR[14] = (unsigned long)adcIRQHandler;
    //Ustaw slot na przerwanie od ADC
    VIC0->VCiR[14] = VIC0_ADC_BIT | VIC_SLOT_EN;
    //Kwalificacja jako przerwanie IRQ
    VIC0->INTSR &= ~VIC0_ADC;
    //Wlaczenie przerwania
    ADC->CR |= ADC_ECVI;
    //Zalaczenie przerwania w VIC
    VIC0->INTER = VIC0_ADC;
    ADC->CR |= ADC_CONT;
}



/*----------------------------------------------------------------------*/

//Funkcja glowna main
int main(void)
{
    //Poczatkowe opoznienie (JTAG CACHE)
    delay(1000000);

    //Ustawienie PLL
    setup_pll();

    //Inicjalizacja portu szeregowego
    uartInit(USED_UART,115200,1,256);

    //Inicjalizacja przetwornika A/C
    setup_adc();

    //Zalacz przerwania
    enable_irq();

    //Napis powitalny
    printf("*****************STR912 Book *********************\n");
    printf("*** Autor: Lucjan Bryndza (c) 2007 Licence GPL ***\n");
    printf("********* ADC z przerwaniami - Woltomierz ********\n");
    printf("**************************************************\n");

    //Start Convert....
    ADC->CR |= ADC_STR;

    //Petla glowna
    while(1)
    {
        printf("ADC=%d\n",adc_val);
	    delay(2000000);
    }
    return 0;
}

