/*
    Str912 book example22 - RTC as periodic interrupt and alarm
    Copyright (C) 2008  Lucjan Bryndza <lucjan.bryndza@ep.com.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "91x_map.h"
#include "91x_scu.h"
#include "91x_fmi.h"
#include <stdint.h>
#include "armint.h"
#include "uart.h"
#include <string.h>
#include <stdlib.h>

/*----------------------------------------------------------------------*/

// delay for n cycles
static void delay(int n)
{
    while(n--)
	asm volatile ("nop");
}

/*----------------------------------------------------------------------*/

/* Ustawienie parametrow zegarowych */
//Czestotliwosc 96MHz
#define PLL_N 192
#define PLL_M 25
#define PLL_P 2
#define SCU_PLLEN 0x80000

static void  setup_pll(void)
{
    //Jezeli petla jest zsynchronizowana tzn ze jest ok
    if(SCU->SYSSTATUS & SCU_FLAG_LOCK) return;
    //Konfiguracja uruchomienie i synchronizacja PLL
    SCU->PLLCONF = (PLL_P<<16) | (PLL_N<<8) | PLL_M;
    SCU->PLLCONF |= SCU_PLLEN;
    while(~SCU->SYSSTATUS & SCU_FLAG_LOCK);

    //Ustawiamy pamiec flash na dwa cykle oczekiwania (dla 96MHz)
    FMI->CR |= FMI_WRITE_WAIT_STATE_1;
    *(vu16*)FMI_BANK_1 = 0x60;
    *(vu16*)(FMI_BANK_1|FMI_READ_WAIT_STATE_2|FMI_LVD_ENABLE|FMI_PWD_ENABLE|FMI_FREQ_HIGH) = 0x03;

    //Ustawiamy F clk na koncu bedzie z PLL-a
    SCU->CLKCNTR = SCU_EMIBCLK_Div2 | SCU_USBCLK_MCLK2 | SCU_PCLK_Div2;
}


/*----------------------------------------------------------------------*/
/* Ustawienie portow IO */
static void setup_gpio(void)
{
    // Wylacz sygnal zerowania dla portu 7 i 3
    SCU->PRR1 |= __GPIO7;

    //Wlacz sygnal zegarowy dla portu 7 i 3
    SCU->PCGR1 |= __GPIO7;

    //Wszystkie wyjscia portu 7 jako push pull
    SCU->GPIOTYPE[7] = 0x0000;

    //Wszystkie wyjscia portu 7 w trybie GPIO
    SCU->GPIOOUT[7]  = 0x5555;

    //Caly port 7 jako wyjsciowy
    GPIO7->DDR = 0xff;
    
}

/*----------------------------------------------------------------------*/
//RTC W enable bit
#define RTC_CR_W (1<<7)

//RTC timer interrupt interval enable
#define RTC_CR_PIE (1<<21)

//RTC alarm interrupt enable
#define RTC_CR_AIE (1<<20)

//RTC alarm enable
#define RTC_CR_AE (1<<20)

//Timer interrupt 128HZ
#define RTC_CR_PISEL_128 (4<<16)

//Timer periodic interrupt flag
#define RTC_SR_PISF (1<<31)

//Timer interrupt alarm flag
#define RTC_SR_AF (1<<30)

////System ticks per second
#define HZ 128


//Bity kontrolera VIC
#define VIC1_RTC (1<<8)
#define VIC1_RTC_BIT 8

//Zalaczenie slotu
#define VIC_SLOTEN (1<<5)


/*----------------------------------------------------------------------*/

//Zmienna odliczajaca
static volatile int tim;

//Flaga alarmu
static volatile int alarm;


//Przerwanie od ukladu czasowo licznikowego
static void irqRTCHandler(void) __attribute__ ((interrupt("IRQ")));

void irqRTCHandler(void)
{
    //Jezeli to przerwanie PERIODIC
    if(RTC->SR & RTC_SR_PISF)
    {
        if(tim) --tim;
    }
    //Jezeli przerwanie od ALARM
    if(RTC->SR & RTC_SR_AF)
    {
        alarm = 1;
    }
    //Zeruj flagi przerwan w RTC
    RTC->SR |= RTC_SR_PISF | RTC_SR_AF;
    //Informacja do VIC - koniec przerwania
    VIC1->VAR = 0;
}

/*----------------------------------------------------------------------*/

enum dayweek {MONDAY=1,TUESDAY,WEDNESDAY,THURSDAY,FRIDAY,SATURDAY,SUNDAY };

struct rtim
{
    uint8_t hour;
    uint8_t minute;
    uint8_t sec;
    uint8_t msec;
    uint16_t year;
    uint8_t month;
    uint8_t day;
    enum dayweek dayw;
};


#define BCD1(x,s)  (((x%10)&0xf)<<s)
#define BCD10(x,s) (((x/10)&0xf)<<s)
#define BCDINT2(x,s) (((x>>(s+4))&0xf)*10 + ((x>>s)&0xf))
#define BCDINT4(x,s) ( ((x>>(s+12))&0xf)*1000+ ((x>>(s+8))&0xf)*100 + ((x>>(s+4))&0xf)*10 + ((x>>s)&0xf))



/* Funkcja ustawia czas zegara RTC */
static void set_rtc_time(const struct rtim *tim)
{
   //Activate write
   RTC->CR |= RTC_CR_W;
   //Setup time
   RTC->TR = BCD10(tim->day,28) | BCD1(tim->day,24) |
             BCD10(tim->hour,20) | BCD1(tim->hour,16) |
             BCD10(tim->minute,12) | BCD1(tim->minute,8) |
             BCD10(tim->sec,4) | BCD1(tim->sec,0);
   //Setup date
   RTC->DTR = BCD10(tim->year/100,28) | BCD1(tim->year/100,24) |
              BCD10(tim->year%100,20) | BCD1(tim->year%100,16) |
              BCD10(tim->month,12) | BCD1(tim->month,8) | (tim->dayw & 0x7);
   //Disable write
   RTC->CR &= ~RTC_CR_W;
}

/*----------------------------------------------------------------------*/
/* Funkcja ustawia czas alarmu */
static void set_rtc_alarm(const struct rtim *tim)
{
   //Activate write
   RTC->CR |= RTC_CR_W;
   //Setup time
   RTC->ATR = BCD10(tim->day,28) | BCD1(tim->day,24) |
             BCD10(tim->hour,20) | BCD1(tim->hour,16) |
             BCD10(tim->minute,12) | BCD1(tim->minute,8) |
             BCD10(tim->sec,4) | BCD1(tim->sec,0);
   //Disable write
   RTC->CR &= ~RTC_CR_W;

  //Alarm enable
  RTC->CR |= RTC_CR_AE;
}

/*----------------------------------------------------------------------*/
static void get_rtc_time(struct rtim *tim)
{
    uint32_t d,t;
    //Zabezpieczenie przed zmiana daty/czasu
    do
    {
        d = RTC->DTR;
        t = RTC->TR;
    }
    while(d != RTC->DTR);
    //Zapis danych struktury
    tim->day = BCDINT2(t,24);
    tim->hour = BCDINT2(t,16);
    tim->minute = BCDINT2(t,8);
    tim->sec = BCDINT2(t,0);
    tim->year = BCDINT4(d,16);
    tim->month = BCDINT2(d,8);
    tim->dayw = BCDINT2(d,0);
}

/*----------------------------------------------------------------------*/
/* Inicjalizacja podstawowych ustawien zegara RTC
 * przerwanie 128Hz oraz przerwanie od alarmu */
static void setup_rtc(void)
{
    /*** Ustawienie Timera 0 w tryb czasomierza ****/
    //Odblokowanie CLK i RST
    SCU->PCGR1 |= __RTC;
    SCU->PRR1  |= __RTC;
    /**** Ustawienia dla RTC *******/
    RTC->CR = RTC_CR_PISEL_128;
    /**** Ustawienie kontrolera przerwan VIC ***/
    //Zalaczenie clocka dla VIC
    SCU->PCGR0 |= __VIC;
    //Reset
    SCU->PRR0 &= ~__VIC;
    SCU->PRR0 |= __VIC;
    //Do wektora podstaw adres procedury obslugi
    VIC1->VAiR[15] = (unsigned long)irqRTCHandler;
    //Ustaw sOClot na przerwanie od WIUOC
    VIC1->VCiR[15] = VIC1_RTC_BIT | VIC_SLOTEN;
    //Zakwalifikuj przerwanie jako IRQ
    VIC1->INTSR &= ~VIC1_RTC;
    //Vic interrupt enable register
    VIC1->INTER = VIC1_RTC;
    //Przerwanie od RTC alarm and periodic
    RTC->CR |= RTC_CR_AIE | RTC_CR_PIE;
}


/*----------------------------------------------------------------------*/
//Wyswietla na terminalu aktualna godzine
static void dispRtim(const struct rtim *tim)
{
    uartWriteUl(UART0,tim->year);
    uartWriteByte(UART0,'-');
    uartWriteUl(UART0,tim->month);
    uartWriteByte(UART0,'-');
    uartWriteUl(UART0,tim->day);
    uartWriteByte(UART0,' ');
    uartWriteUl(UART0,tim->hour);
    uartWriteByte(UART0,':');
    uartWriteUl(UART0,tim->minute);
    uartWriteByte(UART0,':');
    uartWriteUl(UART0,tim->sec);
}

/*----------------------------------------------------------------------*/
//Zdarzenie od timera
static void timer_event(char type)
{
    static uint8_t i;
    // Zalacz wylacz ledy
    if(!alarm)
    {
        GPIO7->DR[0xff<<2]=(1<<i);
    }
    else
    {
        GPIO7->DR[0xff<<2]=0xff;
        //alarm = 0;
	    RTC->SR &= ~RTC_SR_AF;
    }
    if(++i>=8) i=0;
    if(type==0)
    {
	    struct rtim t;
	    get_rtc_time(&t);
	    dispRtim(&t);
	    uartWriteString(UART0,"        \r");
    }
}

/*----------------------------------------------------------------------*/
//Zdarzenie odebranie linii
static void getline_event(char *line,char type)
{
    struct rtim time;
    int tmp[8];
    char *tok = strtok(line,":- ");
    for(int i=0; tok && tim<8; i++)
    {
	tmp[i] = atoi(tok);
	tok = strtok(NULL,":- ");
    }
    time.year = tmp[0];
    time.month = tmp[1];
    time.day = tmp[2];
    time.hour = tmp[3];
    time.minute = tmp[4];
    time.sec = tmp[5];
    if(type=='s')
    {
        set_rtc_time(&time);
    }
    else
    {
        set_rtc_alarm(&time);
        uartWriteString(UART0,"\r\nAlarm o ");
        dispRtim(&time);
        uartWriteString(UART0,"\r\n");
    }
}

/*----------------------------------------------------------------------*/
#define INBUF_SIZE 512

//Bufor znakow
static char inBuf[INBUF_SIZE];

//Pozycja bufora znakow
static uint16_t inBufCnt;


//Funkcja glowna main
int main(void)
{
    
    //Poczatkowe opoznienie (JTAG CACHE)
    delay(1000000);

    //Ustawienie portow gpio
    setup_pll();

    //Ustawienie petli PLL (czestotliwosc 96MHz)
    setup_gpio();

    //Wlaczenie RTC
    setup_rtc();

    //Inicjalizacja portu szeregowego
    uartInit(UART0,115200,1,1024);
    
    //Wlaczenie przerwan
    enable_irq();

    //Napis powitalny
    uartWriteString(UART0,"STR912 Book - RTC.\nCopyright L. Bryndza <lucjan.bryndza@ep.com.pl> GPL v3\n");
    uartWriteString(UART0,"RTC Example\n");
    
    char type = 0;
    // petla nieskocznona
    while(1)
    {
	    int zn = uartReadByte(UART0);
	    //Zdarzenie odebrania linii
	    if(zn!=UART_EMPTY)
	    {
		    uartWriteByte(UART0,zn);
		    if(zn=='\r')
		    {
		        inBuf[inBufCnt] = 0;
		        inBufCnt = 0;
		        uartWriteByte(UART0,'\n');
		        getline_event(inBuf,type);
		        type = 0;
		    }
		    else if(zn=='a' || zn=='s')
		    {
		        uartWriteString(UART0,"Podaj czas?       ");
		        type = zn;
		    }
		    else
		    {
		        //Zapisz znak do bufora
		        if(inBufCnt<INBUF_SIZE) inBuf[inBufCnt++] = zn;
		    }
	    }
	    //Zdarzenie od timera
	    if(!tim)
	    {
		    timer_event(type);
		    tim = HZ/10;
	    }
    }
    return 0;
}

