/*
    Str912 book example25 - Dma memory to memory transfer
    Copyright (C) 2008  Lucjan Bryndza <lucjan.bryndza@ep.com.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "91x_map.h"
#include "91x_scu.h"
#include "91x_fmi.h"
#include "91x_dma.h"
#include "armint.h"
#include "uart.h"
#include <stdio.h>
#include <string.h>
/*----------------------------------------------------------------------*/
#define USED_UART UART0

/*----------------------------------------------------------------------*/

// delay for n cycles
static void delay(int n)
{
    while(n--)
	asm volatile ("nop");
}

/*----------------------------------------------------------------------*/

/* Ustawienie parametrow zegarowych */
//Czestotliwosc 96MHz
#define PLL_N 192
#define PLL_M 25
#define PLL_P 2
#define SCU_PLLEN 0x80000

static void  setup_pll(void)
{
    //Jezeli petla jest zsynchronizowana tzn ze jest ok
    if(SCU->SYSSTATUS & SCU_FLAG_LOCK) return;
    //Konfiguracja uruchomienie i synchronizacja PLL
    SCU->PLLCONF = (PLL_P<<16) | (PLL_N<<8) | PLL_M;
    SCU->PLLCONF |= SCU_PLLEN;
    while(~SCU->SYSSTATUS & SCU_FLAG_LOCK);

    //Ustawiamy pamiec flash na dwa cykle oczekiwania (dla 96MHz)
    FMI->CR |= FMI_WRITE_WAIT_STATE_1;
    *(vu16*)FMI_BANK_1 = 0x60;
    *(vu16*)(FMI_BANK_1|FMI_READ_WAIT_STATE_2|FMI_LVD_ENABLE|FMI_PWD_ENABLE|FMI_FREQ_HIGH) = 0x03;

    //Ustawiamy F clk na koncu bedzie z PLL-a
    SCU->CLKCNTR = SCU_EMIBCLK_Div2 | SCU_USBCLK_MCLK2 | SCU_PCLK_Div2;
}




/*----------------------------------------------------------------------*/
//Memory allocated in FLASH memory
static char src_buf[256] = "Memory to memory DMA transfer for src to dest region. "
			"Cpu is free after dma transfer";

//Memory in SRAM
static char dest_buf[256] = "Another string";

/*----------------------------------------------------------------------*/
#define       DMA_Enable                        0x00000001
#define       DMA_ChannelEnable                 0x00000001
#define       DMA_ChannelDESInc                 0x08000000
#define       DMA_ChannelSRCInc                 0x04000000
#define       DMA_TCIE                          0x80000000

//Enable clocks for DMA controller
static void setup_dma(void)
{
    //Reset kanalu DMA
    SCU->PRR0 &= ~__DMA;
    SCU->PRR0 |= __DMA;

    //Wlacz sygnal zegarowy dla DMA
    SCU->PCGR0 |= __DMA;
 
   //Enable DMA
    DMA->CNFR |= DMA_Enable ;
}

/*----------------------------------------------------------------------*/
//Rozpoczyna kopiowanie danych za pomoca kanalu DMA
static int memcpy_dma_transfer(int channel,char *dest,const char *src,int len)
{
   //DMA Channel table
    static const DMA_Channel_TypeDef *dma_table[] =
    {
        DMA_Channel0, DMA_Channel1, DMA_Channel2, DMA_Channel3,
        DMA_Channel4, DMA_Channel5, DMA_Channel6, DMA_Channel7
    };

    //Channel range check
    if(channel>7) return -1;

    //Current channel table
    DMA_Channel_TypeDef *dma_ch = (DMA_Channel_TypeDef *)dma_table[channel];

    //Clear DMA interrupt
    DMA->TCICR = 1 << channel; 
    DMA->EICR = 1 << channel;

    //Configuration register
    dma_ch->CCNF = DMA_FlowCntrlt0_DMA ; //Memory to memory
    //Source memory area
    dma_ch->SRC = (unsigned long)src;
    //Destination memory area
    dma_ch->DES = (unsigned long)dest;
    //Linked list - not used
    dma_ch->LLI = 0;
    //Configuration register
    dma_ch->CC = DMA_DesWidth_Byte |    //Szerokosc danych wejciowych 1 bajt
                 DMA_SrcWidth_Byte |    //Szerokosc danych wyjsciowych 1 bajt
                 DMA_DesBst_1Data |     //Kopiowanie po 1 bajcie
                 DMA_SrcBst_1Data |     //Kopiowane po 1 bajcie
                 DMA_ChannelSRCInc |    //Dest incr
                 DMA_ChannelDESInc |    //Dest incr
                 DMA_TCIE |             //End of transfer interrupt
                 (len & 0xfff);
    //Enable DMA channel
    dma_ch->CCNF |= DMA_ChannelEnable;
    //Wyjscie z funkcji
    return 0;
}

/*----------------------------------------------------------------------*/
/* Sprawdza status kanalu DMA:
    0 - NIC
    -1 - Blad
    1 - Koniec
*/

static int dma_status(int channel)
{
    //Jezeli blad to -1
    if( DMA->ERISR & (1<<channel) ) return -1;
    //Jezeli koniec transmisji to 1
    if( DMA->TCRISR & (1<<channel) ) return 1;
    return 0;
}

/*----------------------------------------------------------------------*/
//Funkcja glowna main
int main(void)
{
    //Poczatkowe opoznienie (JTAG CACHE)
    delay(1000000);

    //Ustawienie PLL
    setup_pll();

    //Inicjalizacja portu szeregowego
    uartInit(USED_UART,115200,15,128);
	
    //Zalacz przerwania
    enable_irq();
    
    //Konfiguracja DMA
    setup_dma();
    
    //Bufory przed transmisja DMA
    printf("Before dma transfer\nsrc_buf: %s\n",src_buf);
    printf("dest_buf: %s\n",dest_buf);
    
    //Wykonaj transfer DMA
    memcpy_dma_transfer(1,dest_buf,src_buf,strlen(src_buf));

    //Poczekaj na zakonczenie transferu DMA
    while(dma_status(1) == 0)
    {
        //CPU moze tutaj wykonywac dowolne rozkazy, gdy dane sie kopiuja
    }
   
    //Wypisz bufor wynikowy po zakonczeniu transferu DMA
    printf("***AFTER DMA TRANFER***\ndest_buf: %s\n",dest_buf);
    
    return 0;
}


