/*
    Str912 book example3 - simple game gpio based
    Copyright (C) 2007  Lucjan Bryndza <lucjan.bryndza@ep.com.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "91x_map.h"
#include "91x_scu.h"
#include "91x_fmi.h"
#include "gpio.h"

/*----------------------------------------------------------------------*/

// delay for n cycles
static void delay(int n)
{
    while(n--)
	asm volatile ("nop");
}

/*----------------------------------------------------------------------*/

/* Ustawienie parametrow zegarowych */
//Czestotliwosc 96MHz
#define PLL_N 192
#define PLL_M 25
#define PLL_P 2
#define SCU_PLLEN 0x80000

static void  setup_pll(void)
{
    //Jezeli petla jest zsynchronizowana tzn ze jest ok
    if(SCU->SYSSTATUS & SCU_FLAG_LOCK) return;
    //Konfiguracja uruchomienie i synchronizacja PLL
    SCU->PLLCONF = (PLL_P<<16) | (PLL_N<<8) | PLL_M;
    SCU->PLLCONF |= SCU_PLLEN;
    while(~SCU->SYSSTATUS & SCU_FLAG_LOCK);

    //Ustawiamy pamiec flash na dwa cykle oczekiwania (dla 96MHz)
    FMI->CR |= FMI_WRITE_WAIT_STATE_1;
    *(vu16*)FMI_BANK_1 = 0x60;
    *(vu16*)(FMI_BANK_1|FMI_READ_WAIT_STATE_2|FMI_LVD_ENABLE|FMI_PWD_ENABLE|FMI_FREQ_HIGH) = 0x03;

    //Ustawiamy F clk na koncu bedzie z PLL-a
    SCU->CLKCNTR = SCU_EMIBCLK_Div2 | SCU_USBCLK_MCLK2 | SCU_PCLK_Div2;
}



/*----------------------------------------------------------------------*/
//Klasa odczytujaca stan wybranego przycisku w zestawie ZL25ARM
class key
{
public:
        //Numer przycisku
        enum keyNo {key1=0x10,key2=0x20,key3=0x40,key4=0x80};
        //Konstruktor klasy
        key(keyNo id):mGpio(gpioPort::Gpio3),mMask((uint8_t)id) {}
        //Pobierz stan odpowiedniego przycisku
        bool get(void)
        {
            return !mGpio.readIo(mMask);
        }
private:
        //Klasa portu IO
        gpioPort mGpio;
        //Klasa maski bitowej
        uint8_t mMask;

};

/*----------------------------------------------------------------------*/
//Wirtualna klasa bazowa udostepniajaca mozliwosc wyswietlania
class dispObj
{
    public:
        virtual int display(void)=0;
};


/*----------------------------------------------------------------------*/
//Obiekt pilki, ktora moze przemieszczac sie po ekranie
class ball : public dispObj
{
    public:
        //Konstruktor inicjalizujacy
        ball():mPos(1) {}
        //Przesun pilke w prawo
        void stepIn(void)
        {
            mPos<<=1;
            if(mPos==0x100)
            {
                mPos = 0x1;
            }
        }
        //Przesun pilke w lewo
        void stepOut(void)
        {
            if(mPos==1)
            {
                mPos = 0x80;
            }
            else
            {
                mPos>>=1;
            }
        }
        //Wyswietl pozycje pilki
        int display(void) { return mPos; }
        //Czy pilka znajduje sie na koncu w prawo
        bool endIn(void) { return mPos & 0x80; }
        //Czy pilka znajduje sie na koncu w lewo
        bool endOut(void) { return mPos & 0x1; }
    private:
        //Prywatne zawierajace pozycje pilki
        uint16_t mPos;

};

/*----------------------------------------------------------------------*/
// Obiekt wyniku gry dziedziczony od objektu ekranowego
class gameResult: public dispObj
{
    public:
        //Wybor zwyciezcy
        enum eRes {User1=0xC0,User2=0x3};
        //Ustaw zwyciezce
        void set(eRes result) { mResult = result; }
        //Wyswietl zwyciezce
        int display(void) { return mResult; }
    private:
        //Wartosc zwyciezcy
        uint8_t mResult;
};

/*----------------------------------------------------------------------*/
//Obiekt wyswietlacza
class display
{
    public:
        //Konstruktor obiektu
        display():mGpio(gpioPort::Gpio7,0xff) {}
        //Wyswietl obiekt
        void show(dispObj &obj)
        {
            uint8_t v = obj.display();
            mGpio.writeIo(v);
        }
    private:
        //Port IO
        gpioPort mGpio;
};

/*----------------------------------------------------------------------*/
//Klasa gra
class game
{
    public:
        //Konstruktor klasy
        game():mK1(key::key1),mK2(key::key4) {}
        //Funkcja stepujaca wywolywana z main
        void step(void);
    private:
        gameResult mResult;     //Wynik gry
        ball mBall;             //Pilka
        key mK1,mK2;            //Przyciski
        display mDisp;
        enum gameState {toOut,toIn,dispWin1,dispWin2 } mState;  //Stan gry
        //Przesuwanie polki w prawo
        void stepIn();
        //Przesywanie pilki w lewo
        void stepOut();
        //Wyswietlanie rezultatu
        void stepResult();
};

//Prywatna metoda wywolywana gdy pilka porusza sie w prawo
void game::stepIn()
{
        //Przesun pilke o jedna pozycje w prawo
        mBall.stepIn();
        //Czy pilka na brzegu prawej krawdzi
        if(mBall.endIn())
        {
            //Czy klawisz wcisniety
            if(mK2.get())
            {
               //Tak - pilka odbija sie w strone przeciwnika
                mState = toOut;
            }
            else
            {
                //Nie - przegrales wyswietlany jest zwyciezca
                mState = dispWin1;
            }
        }
        //Pokaz pilke
        mDisp.show(mBall);
}


//Prywatna metoda wywolywana gdy pilka porusza sie w lewo
void game::stepOut()
{
        //Przesun pilke w lewo
        mBall.stepOut();
        //Czy pilka znajduje sie w lewej krawedzi?
        if(mBall.endOut())
        {
            //Czy wcisniety lewy przycisk
            if(mK1.get())
            {
               //Tak - odbijaj pilke w prawo
                mState = toIn;
            }
            else
            {
                //Nie - przegrales wyswietl wynik
                mState = dispWin2;
            }
        }
        //Pokaz pilke
        mDisp.show(mBall);
}

//Prywatna metoda wyswietlajaca wynik rozgrywek
void game::stepResult()
{
    //Sprawdz kto wygral i ustaw tablice wynikow
    if(mState==dispWin1) mResult.set(gameResult::User1);
    else mResult.set(gameResult::User2);
    //Wyswietl tabice wynikow
    mDisp.show(mResult);
    //Jezeli ktorys klawisz wcisniety to rozpocznik gre od nowa
    if(mK1.get() || mK2.get())
    {
        mState = toIn;
    }
}


//Metoda wywolywana z funkcji glownej
void game::step(void)
{
    //Sprawdz stan
    switch(mState)
    {
        //Jezeli w prawo wywolaj metode obslugi poruszania sie w prawo
        case toOut:
            stepOut();
            break;
        //Jezeli w lewo wywolaj metode poruszania sie w lewo
        case toIn:
            stepIn();
            break;
       //Jezeli przgrales wywolaj metode wyswietlania wynikow
       case dispWin1:
       case dispWin2:
            stepResult();
            break;
    }
}

/*----------------------------------------------------------------------*/
//Obiekt glowny gry...
game MyGame;

//Funkcja glowna (main)
int main(void)
{
    //Poczatkowe opoznienie (Awaryjny JTAG)
    delay(1000000);

    //Ustawienie portow gpio
    setup_pll();

    // petla nieskocznona
    while(1)
    {
	    //Krok gry
        MyGame.step();
        //Opoznienie
        delay(3000000);
    }
    return 0;
}

